/**
 * P4
 * http://snarfed.org/space/p4
 *
 * Copyright 2001-2004
 * Ken Ashcraft <kashcraft+p4@gmail.com>
 * Ryan Barrett <p4@ryanb.org>
 * Maulik Shah <p4@maulik.net>
 * Nathan Stoll <nathanstoll+p4@gmail.com>
 *
 * @file conn.h
 *
 * Defines the cconn class, which is the object for an open connection to
 * another P4 client. There are methods for sending and receiving data, and the
 * ctor and dtor open and close the connection, respectively.
 *
 * The ctor(s) are protected so that only the cconn_mgr class can create and
 * destroy cconns. The cconn_mgr class is an Abstract Factory for cconn, but
 * with other functionality as well.
 *
 * TODO: Store cconn *'s in list instead of cconn, so copy ctor can be private.
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
 * more details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA 02111-1307 USA
 */

#ifndef _CONN_H
#define _CONN_H

#include "protocol.h"
#include <iostream>
#include <string>
using namespace std;


/*****************************************************************************
 * FORWARD DECLARATIONS
 *****************************************************************************/
class cconn_mgr;  // so that we can declare it as a friend
class cmsg;


/*****************************************************************************
 * CCONN CLASS
 *****************************************************************************/
class cconn {
  public:
  /** copy ctor
   *
   * Public so that the list can store it.
   */
  cconn(const cconn &);

  /** dtor
   *
   * Destroys the cconn, and disconnects if connected.
   */
  ~cconn();

  /** operator <<
   *
   * The stream insertion operator for cconn - prints out vital information
   * about the connection.
   *
   * NOTE: this is a *non*-member function so that the ostream can be the lhs
   * of the operator<<.
   */
  friend ostream &operator<< (ostream &ostr, const cconn &conn);

  /** Type: estatus_t
   *
   * Enumerated type for the status of a cconn. Somewhat self-explanatory.
   *
   *  eopening:
   * Connector-side only. The TCP connection has been opened, but it isn't
   * established yet.
   *
   *  epreamble_wait:  
   * Connectee-side only. The TCP connection has been established, but the P4
   * preamble ("P4 CONNECT\<version number>") hasn't been received yet.
   *
   *  epreamble_ack_wait:
   * Connector-side only. The P4 preamble has been sent, but the preamble ack
   * ("P4 OK") hasn't been received yet.
   *
   *  econnected:
   * Both sides. The preamble has been completed and the connection is ready
   * to send and receive data.
   */
  typedef enum estatus_tt {
    eopening, epreamble_wait, epreamble_ack_wait, econnected, efailed, eclosed
  } estatus_t;

  estatus_t get_status() const { return status; }

  /** send
   *
   * Sends the given message over this connection.
   *
   * Returns kerr_success if the message was sent, otherwise returns the
   * specific message.
   *
   * If the message is in host byte order, send converts it to network byt
   * order.
   */
  uint send(cmsg &msg);

  /** wait_until_ready
   *
   * Blocks the calling thread (efficiently) until the connection has
   * completed the preamble and is ready to send and receive data.
   *
   * Returns kerr_success if the socket is ready, otherwise returns the
   * specific error.
   */
  uint wait_until_ready() const;


  protected:
  friend class cconn_mgr;      // so that cconn_mgr can manage cconns!
  friend class cmsg;           // so that cmsg can call send_all
  friend class cmsg_discover;  // to access failed instance

  /** operator==
   *
   * Two connections are equal if they have the same socket.
   *
   * NOT USED right now - no need for it. Might be useful in the future.
   */
  // bool operator==(const cconn &rhs) { return (sock == rhs.sock); }

  /** get_socket
   *
   * Returns this connection's socket.
   */
  int get_socket() const    { return sock; }

  /** is_permanent
   *
   * Accessors for the permanent member. If a connection is permanent, it was
   * opened with the CONNECT message. Otherwise, it was opened with another
   * message and the CONNECT message has not been sent or received.
   *
   * NOTE: currently, transient connections are closed immediately after they
   * finish sending or receiving.
   */
  bool is_permanent() const { return permanent; }
  void make_permanent() { permanent = true; }

  /** send_all
   *
   * Calls (socket) send repeatedly until the given buffer has been completely
   * sent. If send fails, send_all returns kerr_not_sent.
   */
  uint send_all(const void *buf, uint size) const;

  /** receive
   *
   * Receives data over this connection. Once the header has been received,
   * a new cmsg object is created (using cmsg_factory), and following calls to
   * receive will receive into the message's data buffer. Once all of the data
   * has been received, process() will be called on the message.
   *
   * NOTE: This call is *blocking!!!* - it should only be called (by
   * cconn_mgr) after it has determined by select() that there is data waiting
   * to be read.
   *
   * Returns kerr_success on success, otherwise returns the specific error.
   */
  uint receive();

  /** close
   *
   * Closes the socket and sets the connection's status to eclosed.
   */
  void close();

  /** is_idle
   *
   * Returns true if the connection is not currently receiving a message,
   * false otherwise.
   *
   * NOTE: If the connection is not ready - ie if it is being established or
   * in the middle of the preamble - is_idle returns false.
   */
  bool is_idle() const;

  /** failed_instance
   *
   * Returns an instance of cconn with status efailed. (Useful in
   * cconn_mgr and other places that return a cconn, but may hit an
   * error.)
   */
  static cconn &failed_instance();
  

  /** ctors
   *
   * Creates a new cconn - can only be called by cconn_mgr. The first ctor
   * opens a connection to the given address and port, and the second ctor
   * accepts a socket that has already been opened. In either case, the conn
   * handles the preamble itself.
   */
  cconn(ulong addr, ushort port, const cmsg &init_msg, const char *username);
  cconn(int sock, bool failed = false);


  private:
  /** recv_preamble
   *
   * If the connection hasn't been completely established yet, this receives
   * data on the socket and contributes it to the preamble (and updates the
   * connection's status, if appropriate).
   */
  uint recv_preamble();

  /**************************************************************************
   * CONSTANTS AND MEMBERS
   **************************************************************************/
  protected:
  static const uint kconn_wait_until_ready_timeout_msecs;

  int sock;
  string peer_username;
  bool permanent;

  estatus_t status;

  char preamble_buf[sizeof(kp4_preamble)];
  uint preamble_bytes_read;

  p4_header hdr;
  uint hdr_bytes_left;
  cmsg *msg;
};




/*****************************************************************************
 * INLINE METHODS
 *****************************************************************************/
inline cconn &cconn::failed_instance()
{
  static cconn failed(-1, true);  // give it -1 as the socket

  return failed;
}


inline bool cconn::is_idle() const
{
  // only idle if not receiving message *and* have completed preamble
  return (status == econnected && hdr_bytes_left == sizeof(hdr));
}


#endif  // _CONN_H
